import click
import logging
from .configure import CondaInstaller, CONDA, SystemNotReady, SystemNotSupported, InstallFailed, ShellNotSupported
from ..utils import RunError


logger = logging.getLogger(__name__)


def cleanup_install():
    CondaInstaller.kill()


@click.command(help="Configure system to run Aero")
def configure():
    click.echo('Checking system for conda...')
    if CondaInstaller.check_installed(CONDA):
        click.echo('conda installed, system configured')
        return
    else:
        click.echo('conda not installed, attempting conda install')
        inst = CondaInstaller()

        try:
            ret = inst.install_conda()
            if ret != 0:
                click.echo(f'Sorry, there was an error')
            else:
                click.echo(f'Configuration complete, please close shell and reopen')

        except SystemNotSupported as e:
            click.echo(f'Configure command does not support {e.system} {e.machine}')
            click.echo(f'Please install conda manually.')

        except SystemNotReady as e:
            click.echo(f'System does not have {e.missing}, please install this for configure to work')

        except InstallFailed as e:
            click.echo(f'Install failed: {e.cmd} returned with error {e.ret}')

        except ShellNotSupported as e:
            click.echo(f'Configure command does not support shell: {e.shell}')
            click.echo(f'Please install conda manually.')



import click
from ..utils import red, RunError
import logging
import os
from ..run.commands import cmd_run
from ..run.run import AeroRun, RunState, FileNotSupported, BATCH_DEC, RESOURCE_DEC

logger = logging.getLogger(__name__)


@click.group(help="Resume a flow")
def resume():
    pass


@resume.command(help="Resume a flow locally")
@click.argument('filename')
@click.option("--step", type=str, default=None)
@click.option("--origin-run-id", type=str, default=None)
@click.option("--args", default=None,
              help="Flow arguments, must be presented as a quoted string, eg "
                   "'--arg1 arg1 --arg2 arg2'")
def local(filename, step, origin_run_id, args):
    cmd_run(filename, args, RunState.LOCAL, resume=True, step=step, from_run_id=origin_run_id)


@resume.command(help="Resume a flow in the cloud")
@click.argument('filename')
@click.option("--step", type=str, default=None)
@click.option("--origin-run-id", type=str, default=None)
@click.option("--cpu", type=int, default=None)
@click.option("--memory", type=int, default=None)
@click.option("--args", default=None,
              help="Flow arguments, must be presented as a quoted string, eg "
                   "'--arg1 arg1 --arg2 arg2'")
def cloud(filename, step, origin_run_id, cpu, memory, args):
    cmd_run(filename, args, RunState.CLOUD, cpu, memory, resume=True, step=step, from_run_id=origin_run_id)



import click
from ..utils import red, SITE_BASE, print_welcome, cyan
from .account import Account, UserNotFound, IncorrectPassword, ProvisionError, ComputeStatus, FreeTrialEnded
import webbrowser
import logging
import time
from rich.console import Console

logger = logging.getLogger(__name__)


@click.group(help="Account management")
def account():
    pass


@account.command(help="Create an account")
def create():

    print_welcome()

    click.echo(f'CLI does not currently support Account creation.')
    click.echo(f'Please create an account on our website.')
    webbrowser.open(f'{SITE_BASE}/register')
    return


@account.command(help="Login to an account")
@click.option("--email", default=None,
              help="Email address")
@click.option("--password", default=None,
              help="Password")
def login(email, password):

    print_welcome()

    if not email:
        email = click.prompt("Email > ")

    if not password:
        password = click.prompt("Password > ", hide_input=True)

    try:
        user_account = Account(email, password)
        user_account.login()

    except UserNotFound:
        click.echo(red("User not found, check username"))
        return
    except IncorrectPassword:
        click.echo(red("Password not recognised"))
        return
    except FreeTrialEnded:
        click.echo(red("*** Your free trial of Aero has ended ***"))
        click.echo("If you're interesting in using our platform more, please sign up to hear from us: " + cyan("https://site.aeroplatform.co.uk"))
        return
    except Exception as e:
        click.echo(
            red("An error occured when attempting to log in, please try again"))
        return

    try:
        provision(user_account)

    except Exception as e:
        click.echo(
            "An error occured when attempting to provision, please try again")
        logger.debug(e)

    return


def provision(user_account):

    status = ComputeStatus.NO_VALUE
    is_first_provision = False

    console = Console()

    with console.status("[bold green]Logging in...") as status:
        while status != ComputeStatus.CREATED:
            status = user_account.provision(is_first_provision)
            logger.debug(f"Current Status is {status}")

            if status == ComputeStatus.INIT:
                logger.debug(f"Initialised Compute Environment Creation")
                console.print(
                    "Initialising Compute Environment, this might take a few minutes!")
                is_first_provision = True

            if status == ComputeStatus.CREATING:
                logger.debug(f"Creating Compute Environment...")
                is_first_provision = True

            if status == ComputeStatus.CREATED:
                break

            time.sleep(30)

        click.echo("")
        click.echo("Compute Environment Created")
        click.echo(
            f"You are now ready to run Flows with Aero. For some basic tutorials, visit: {SITE_BASE}/commands")



import click
from ..utils import red, check_file_exists, RunError
import logging
import os
from .run import AeroRun, RunState, FileNotSupported, BATCH_DEC, RESOURCE_DEC

logger = logging.getLogger(__name__)


def cleanup_run():
    AeroRun.kill()


@click.group(help="Run flows")
def run():
    pass


@run.command(help="Run a flow locally")
@click.argument('filename')
@click.option("--args", default=None,
              help="Flow arguments, must be presented as a quoted string, eg "
                   "'--arg1 arg1 --arg2 arg2'")
def local(filename, args):
    cmd_run(filename, args, RunState.LOCAL)


@run.command(help="Run a flow in the cloud")
@click.argument('filename')
@click.option("--cpu", type=int, default=None)
@click.option("--memory", type=int, default=None)
@click.option("--args", default=None,
              help="Flow arguments, must be presented as a quoted string, eg "
                   "'--arg1 arg1 --arg2 arg2'")
def cloud(filename, cpu, memory, args):
    cmd_run(filename, args, RunState.CLOUD, cpu, memory)


def cmd_run(filename: str, args: str, run_state: RunState, cpu: int = None, mem: int = None,
            resume: bool = False, step: str = None, from_run_id: str = None):
    try:
        if check_file_exists(filename):                
            flow = AeroRun(filename, args, run_state, cpu, mem, resume, step, from_run_id)

            if flow.check_if_local_and_has_batch_decorator():

                warning = f'Warning, {BATCH_DEC} decorator detected in file and running local, please note some ' \
                            f'steps will use the cloud - do you want to continue?'
                if not click.confirm(warning):
                    click.echo(red('Aborting...'))
                    return

            if flow.check_if_local_and_has_resources_decorator():
                warning = f'Requested resources using {RESOURCE_DEC} may not match available system resources ' \
                            f'- flow may fail, do you want to continue?'
                if not click.confirm(warning):
                    click.echo(red('Aborting...'))
                    return

            r = flow.run()

            if r > 0:
                click.echo(f'Flow completed but there may have been errors: {r}')

        else:
            click.echo(red('File not found.'))
            return

    except FileNotSupported as e:
        click.echo(f'{e.filename}: {red(e.message)}!')
    except RunError:
        click.echo('There was an error running the flow.')



import platform
import logging
from ..utils import AeroProcessRunner, get_shell, get_supported_shells
from shutil import which
from typing import List
import os
from enum import Enum, unique

logger = logging.getLogger(__name__)

CONDA = 'conda'
CURL = 'curl'
WGET = 'wget'
DARWIN = 'Darwin'
MACOSX = 'MacOSX'
LINUX = 'Linux'
SYSTEMS = [DARWIN, LINUX]
MACHINES = ['x86', 'x86_64', 'armv7l', 'aarch64']
MINICONDA_SCRIPT_PREFACE = 'Miniconda3-latest'
MINICONDA_SCRIPT_LOCALE = f"{os.environ['HOME']}/tmp_aero/"
MINICONDA_INSTALL_LOCALE = f"{os.environ['HOME']}/miniconda"
MINICONDA_SCRIPT_NAME = 'miniconda.sh'
CONDA_URL_BASE = 'https://repo.anaconda.com/miniconda/'


class SystemNotSupported(Exception):

    def __init__(self, machine: str, system: str):
        super().__init__()
        self.machine = machine
        self.system = system


class SystemNotReady(Exception):

    def __init__(self, cmd: str):
        super().__init__()
        self.missing = cmd


class ShellNotSupported(Exception):

    def __init__(self, shell: str):
        super().__init__()
        self.shell = shell


class InstallFailed(Exception):

    def __init__(self, cmd: str, ret: int):
        super().__init__()
        self.cmd = cmd
        self.ret = ret


@unique
class InstallOrder(Enum):
    __order__ = 'PRE_INSTALL DOWNLOAD INSTALL CLEAN CONFIG INIT'
    PRE_INSTALL = 0
    DOWNLOAD = 1
    INSTALL = 2
    CLEAN = 3
    CONFIG = 4
    INIT = 5


class CondaInstaller(AeroProcessRunner):

    def __init__(self):
        super().__init__()
        self.system = None
        self.machine = None
        self.conda_script_name = None
        self._commands = [str() for s in InstallOrder]

    def install_conda(self):
        self.system = platform.system()
        self.machine = platform.machine()

        if self.system not in SYSTEMS or \
           self.machine not in MACHINES:
            raise SystemNotSupported(machine=self.machine, system=self.system)

        if self.system == DARWIN:
            self.system = MACOSX

        self.conda_script_name = f'{MINICONDA_SCRIPT_PREFACE}-{self.system}-{self.machine}.sh'
        logger.debug(f'Conda Target Script: {self.conda_script_name}')

        self._commands[InstallOrder.PRE_INSTALL.value] = ['mkdir', MINICONDA_SCRIPT_LOCALE]

        if self.system == MACOSX:
            self._commands[InstallOrder.DOWNLOAD.value] = self._get_install_conda_cmd_darwin(self.conda_script_name)
        elif self.system == LINUX:
            self._commands[InstallOrder.DOWNLOAD.value] = self._get_install_conda_cmd_linux(self.conda_script_name)

        self._commands[InstallOrder.INSTALL.value] = ['bash',
                                                      f'{MINICONDA_SCRIPT_LOCALE}{MINICONDA_SCRIPT_NAME}',
                                                      '-b', '-p', MINICONDA_INSTALL_LOCALE]

        self._commands[InstallOrder.CLEAN.value] = ['rm', '-rf', MINICONDA_SCRIPT_LOCALE]

        self._commands[InstallOrder.CONFIG.value] = [f'{MINICONDA_INSTALL_LOCALE}/bin/conda', 'config', '--set',
                                                     'auto_activate_base', 'false']
        shell = get_shell()
        if shell in get_supported_shells():
            self._commands[InstallOrder.INIT.value] = [f'{MINICONDA_INSTALL_LOCALE}/bin/conda', 'init', shell]
            return self._install_conda()
        else:
            raise ShellNotSupported(shell)

    def _install_conda(self):
        ret = 0
        for command_idx in InstallOrder:
            ret = self._run_blocking_process(cmd=self._commands[command_idx.value])
            if ret != 0:
                raise InstallFailed(cmd=self._commands[command_idx.value], ret=ret)

        return ret

    @classmethod
    def check_installed(cls, cmd: str) -> bool:
        return which(cmd) is not None

    @staticmethod
    def _get_install_conda_cmd_darwin(script_name: str) -> List:
        if CondaInstaller.check_installed(CURL):
            return [CURL, f'{CONDA_URL_BASE}{script_name}', '-o', f'{MINICONDA_SCRIPT_LOCALE}{MINICONDA_SCRIPT_NAME}']
        else:
            raise SystemNotReady(cmd=CURL)

    @staticmethod
    def _get_install_conda_cmd_linux(script_name: str) -> List:
        if CondaInstaller.check_installed(WGET):
            return [WGET, f'{CONDA_URL_BASE}{script_name}', '-O', f'{MINICONDA_SCRIPT_LOCALE}{MINICONDA_SCRIPT_NAME}']
        else:
            raise SystemNotReady(cmd=WGET)



from enum import Enum
import click
import os
import re
import subprocess
import logging
from typing import List

logger = logging.getLogger(__name__)

def cyan(string: str):
    return click.style(string, fg='cyan')


def magenta(string: str):
    return click.style(string, fg='magenta')


def red(string: str):
    return click.style(string, fg='red')


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
SITE_BASE = "https://site.aeroplatform.co.uk"
TUTORIAL_PATH = "/commands"
NAME_LOWER = "aero"
NAME_UPPER = "Aero"
NAME_ASCII = cyan('\n'
                  ',adPPYYba,   ,adPPYba,  8b,dPPYba,   ,adPPYba, \n'  
                  '""     `Y8  a8P_____88  88P    "Y8  a8"     "8a\n'  
                  ',adPPPPP88  8PP"""""""  88          8b       d8\n'  
                  '88,    ,88  "8b,   ,aa  88          "8a,   ,a8"\n'  
                  '`"8bbdP"Y8   `"Ybbd8"   88           `"YbbdP"  \n')


def print_welcome():
    click.echo(NAME_ASCII)
    click.echo(f'Welcome to {magenta(NAME_UPPER)}!')
    click.echo("A Data Platform designed to put developers first.")
    click.echo(f"For more information, visit {cyan(SITE_BASE)}")


def check_file_exists(filename: str) -> bool:
    return os.path.isfile(filename)

class Executable(Enum):
    PYTHON = "python3"
    R = "Rscript"

class DecoratorNotFoundException(Exception):

    def __init__(self, message: str, decorator: str):
        super().__init__(message)
        self.message = message
        self.decorator = decorator

class DecoratorFormatter():

    format = {}
    decorator = "base"

    def __init__(self, executable: Executable):
        self._executable = executable

    def in_line(self, line: str):
        
        # If None, decorator not in line
        return re.search(self.format[self._executable], line) != None

    def raise_not_found(self):
        raise DecoratorNotFoundException(
            f"Decorator not found, please check documentation at {SITE_BASE}{TUTORIAL_PATH}", 
            self.decorator
        )

class CondaDecoratorFormat(DecoratorFormatter):

    format = {
        Executable.PYTHON: "@conda",
        Executable.R: r"\A(?!x)x" # a regex which can never match, R doesn't need --env=conda
    }
    decorator = "Conda"

def get_shell() -> str:
    try:
        return os.path.basename(os.environ['SHELL'])
    except KeyError:
        return ''


def get_supported_shells() -> List:
    return ['bash', 'fish', 'zsh', 'tcsh']


class FileNotSupported(Exception):

    def __init__(self, message: str, filename: str):
        super().__init__(message)
        self.message = message
        self.filename = filename


class RunError(Exception):
    ...


class AeroProcessRunner:

    ACTIVE_RUNS = []

    def __init__(self):
        self._sp = None

    def _run_blocking_process(self, cmd: List) -> int:
        logger.debug(f'Running blocking command: {cmd}')

        try:
            self._sp = subprocess.Popen(cmd)
        except Exception as e:
            logger.debug(f'Exception from Popen: {e}')
            raise RunError

        self._add_active_proc(self._sp)
        self._sp.wait()
        self._rem_active_proc(self._sp)

        return self._sp.returncode

    @classmethod
    def _add_active_proc(cls, proc: subprocess.Popen):
        cls.ACTIVE_RUNS.append(proc)

    @classmethod
    def _rem_active_proc(cls, proc: subprocess.Popen):
        try:
            cls.ACTIVE_RUNS.remove(proc)
        except ValueError:
            logger.debug("Tried to remove a proc that was not added")

    @classmethod
    def kill(cls):
        for proc in cls.ACTIVE_RUNS:
            logger.debug(f'Killing active process {proc}')
            proc.kill()

    @classmethod
    def get_executable(cls, filename: str) -> Executable:

        # Catching error like this allows us to not have a NOT_SUPPORTED executable type,
        # Else for the function to return an executable, we'd have to have a "None" enum
        try:
            _, file_extension = os.path.splitext(filename.lower())
            extensions = {
                ".py": Executable.PYTHON,
                ".r": Executable.R
            }

            return extensions[file_extension]
        except:
            raise FileNotSupported('Currently, only Python and R files are supported', filename)

    @classmethod
    def check_for_decorator(cls, filename: str, dec: DecoratorFormatter) -> bool:
        with open(filename) as f:
            for line in f:
                if dec.in_line(line):
                    li = line.strip()
                    if not li.startswith('#'):
                        logger.debug(f'{dec} is set')
                        return True
        return False



import click
import logging
from ..utils import DecoratorNotFoundException, check_file_exists, red, FileNotSupported, RunError
from .schedule import AeroScheduler, ScheduleDecoratorFormat


logger = logging.getLogger(__name__)


def cleanup_scheduler():
    AeroScheduler.kill()


@click.group(help="Schedule flows and inspect scheduled flows")
def schedule():
    pass


@schedule.command(help="Create a new scheduled Flow")
@click.argument('filename')
def create(filename):
    scheduler = _init_scheduler(filename)
    try:
        if scheduler:
            r = scheduler.create()
            if r > 0:
                click.echo(f'Task completed but there may have been errors: {r}')
        else:
            return

    except RunError:
        click.echo('There was an error creating the flow schedule.')


@schedule.command(help="View all runs of a Flow")
@click.argument('filename')
def inspect(filename):
    scheduler = _init_scheduler(filename)
    try:
        if scheduler:
            r = scheduler.inspect()
            if r > 0:
                click.echo(f'Task completed but there may have been errors: {r}')
        else:
            return

    except RunError:
        click.echo('There was an error inspecting the flow schedule.')


@schedule.command(help="Trigger a scheduled Flow to run immediately")
@click.argument('filename')
def trigger(filename):
    scheduler = _init_scheduler(filename)
    try:
        if scheduler:
            r = scheduler.trigger()
            if r > 0:
                click.echo(f'Task completed but there may have been errors: {r}')
        else:
            return

    except RunError:
        click.echo('There was an error triggering the scheduled flow.')


@schedule.command(help="View logs from a Flow")
@click.argument('filename')
def logs(filename):
    scheduler = _init_scheduler(filename)
    try:
        if scheduler:
            r = scheduler.logs()
            if r > 0:
                click.echo(f'Task completed but there may have been errors: {r}')
        else:
            return

    except RunError:
        click.echo('There was an error viewing logs for the schedule.')


@schedule.command(help="Remove a scheduled flow")
@click.argument('filename')
def remove(filename):
    scheduler = _init_scheduler(filename)
    try:
        if scheduler:
            r = scheduler.remove()
            if r > 0:
                click.echo(f'Task completed but there may have been errors: {r}')
        else:
            return

    except RunError:
        click.echo('There was an error removing the schedule.')


def _init_scheduler(filename: str) -> AeroScheduler:
    try:
        if not _check_file(filename):
            return None
        else:
            scheduler = AeroScheduler(filename)
            return scheduler

    except FileNotSupported as e:
        click.echo(f'{e.filename}: {red(e.message)}!')
    except DecoratorNotFoundException as e:
        click.echo(f'{e.decorator}: {red(e.message)}!')

    return None


def _check_file(filename: str) -> bool:
    if check_file_exists(filename):
        return True
    else:
        click.echo(red('File not found.'))

    return False



from ..utils import AeroProcessRunner, DecoratorFormatter, Executable, CondaDecoratorFormat
from ..run.run import CONDA_ENV
import subprocess
import logging

logger = logging.getLogger(__name__)

class ScheduleDecoratorFormat(DecoratorFormatter):

    format = {
        Executable.PYTHON: "@schedule",
        Executable.R: "decorator.*schedule"
    }
    decorator = "Schedule"

SCHEDULE_CMD = 'schedule'
SCHEDULE_CMD_CREATE = 'create'
SCHEDULE_CMD_LOGS = 'get-logs'
SCHEDULE_CMD_INSPECT = 'list-runs'
SCHEDULE_CMD_TRIGGER = 'trigger'
SCHEDULE_REMOVE_FLOW = "remove"


class AeroScheduler(AeroProcessRunner):

    def __init__(self, filename: str):
        super().__init__()
        self._filename = filename
        
        # Moved decorator check in ProcessRunner so we can 
        # base decorator checks based on executable
        decorator = ScheduleDecoratorFormat(self.get_executable(filename))

        if not self.check_for_decorator(
            filename, 
            decorator
        ):  
            # raises an exception
            decorator.raise_not_found()

    def create(self):
        return self._scheduler_run(SCHEDULE_CMD_CREATE)

    def trigger(self):
        return self._scheduler_run(SCHEDULE_CMD_TRIGGER)

    def inspect(self):
        return self._scheduler_run(SCHEDULE_CMD_INSPECT)

    def logs(self):
        return self._scheduler_run(SCHEDULE_CMD_LOGS)

    def remove(self):
        return self._scheduler_run(SCHEDULE_REMOVE_FLOW)

    def _scheduler_run(self, command: str):
        # Build command
        executable = self.get_executable(self._filename)
        aero_cmd = [executable.value, self._filename]

        if self.check_for_decorator(self._filename, CondaDecoratorFormat(executable)):
            aero_cmd.append(CONDA_ENV)

        aero_cmd.extend([SCHEDULE_CMD, command])

        return self._run_blocking_process(aero_cmd)



import click
import os
import logging
import sys
from .account.commands import account
from .run.commands import run, cleanup_run
from .schedule.commands import schedule, cleanup_scheduler
from .resume.commands import resume
from .configure.commands import configure, cleanup_install
import signal

logging.getLogger('boto3').setLevel(logging.CRITICAL)
logging.getLogger('botocore').setLevel(logging.CRITICAL)
logging.getLogger('s3transfer').setLevel(logging.CRITICAL)
logging.getLogger('urllib3').setLevel(logging.CRITICAL)

logger = logging.getLogger(__name__)


def signal_handler(sig, frame):
    logger.debug("SIGINT detected, cleaning up active runs")
    cleanup_run()
    cleanup_scheduler()
    cleanup_install()
    sys.exit(0)


@click.group()
def cli():
    logging.basicConfig(stream=sys.stdout, level=os.environ.get('LOGLEVEL', 'INFO').upper(),
                        format='%(filename)s %(lineno)d: %(message)s')
    signal.signal(signal.SIGINT, signal_handler)
    pass


cli.add_command(account)
cli.add_command(run)
cli.add_command(resume)
cli.add_command(schedule)
cli.add_command(configure)


if __name__ == '__main__':
    cli()



import os
import logging
from enum import Enum, unique
import subprocess
import re
import os.path
import psutil
import re
from ..utils import AeroProcessRunner, FileNotSupported, \
    Executable, DecoratorFormatter, CondaDecoratorFormat

logger = logging.getLogger(__name__)

class ResourceDecoratorFormat(DecoratorFormatter):

    format = {
        Executable.PYTHON: "@resources",
        Executable.R: "decorator.*resources"
    }
    resource_format = {
        Executable.PYTHON: f"@resources\(([^)]+)\)",
        Executable.R: "decorator\([^=]*,([^)]+)\)"
    }
    decorator = "Resources"

    def parse_resources(self, line):
        return re.search(self.resource_format[self._executable], line)


class BatchDecoratorForamt(DecoratorFormatter):

    format = {
        Executable.PYTHON: "@batch",
        Executable.R: "decorator.*batch"
    }
    decorator = "Resources"

BATCH_DEC = '@batch'
RESOURCE_DEC = '@resources'
CONDA_ENV = '--environment=conda'
MF_RUN = 'run'
MF_RESUME = 'resume'
BATCH_CMD = '--with batch'
RUN_ID_OPTION = "--origin-run-id"


@unique
class RunState(Enum):
    LOCAL = 0
    CLOUD = 1


class AeroRun(AeroProcessRunner):

    def __init__(self, filename: str, args: str, run_state: RunState, cpu: int = None, mem: int = None,
                 resume: bool = False, step: str = None, from_run_id: str = None):
        super().__init__()
        self._filename = filename
        self._args = args
        self._run_state = run_state
        self._cpu = cpu
        self._mem = mem
        self._resume = resume
        self._step = step
        self._from_run_id = from_run_id
        self._executable = self.get_executable(self._filename)

    def check_if_local_and_has_batch_decorator(self):
        batch_decorator = BatchDecoratorForamt(self._executable)
        return self._run_state == RunState.LOCAL and \
            AeroRun.check_for_decorator(self._filename, batch_decorator)

    def check_if_local_and_has_resources_decorator(self):
        return self._run_state == RunState.LOCAL and \
            not self.check_resources()

    def run(self):
        # Build command
        aero_cmd = [self._executable.value, self._filename]

        # Check if conda decorator was found, if so, add env
        if self.check_for_decorator(self._filename, CondaDecoratorFormat(self._executable)):
            aero_cmd.append(CONDA_ENV)

        if not self._resume:
            # Add metaflow run command
            aero_cmd.append(MF_RUN)
        else:
            # Add metaflow resume command
            aero_cmd.append(MF_RESUME)

            if self._step:
                aero_cmd.append(self._step)

            if self._from_run_id:
                aero_cmd.append(RUN_ID_OPTION)
                aero_cmd.append(self._from_run_id)

        # Add any user args
        if self._args:
            aero_cmd.extend(self._args.split(" "))

        # Check if run is cloud, if so check resources
        if self._run_state == RunState.CLOUD:
            cloud_cmd = BATCH_CMD
            if self._cpu or self._mem:
                cloud_cmd += ':'
                if self._cpu:
                    cloud_cmd += f'cpu={self._cpu},' if self._mem else f'cpu={self._cpu}'
                if self._mem:
                    cloud_cmd += f'memory={self._mem}'

            aero_cmd.extend(cloud_cmd.split(" "))

        return self._run_blocking_process(aero_cmd)

    def check_resources(self) -> bool:
        memory = 0
        cpu = 0
        check_resources = False
        resource_decorator = ResourceDecoratorFormat(self._executable)

        with open(self._filename) as f:
            for line in f:
                if resource_decorator.in_line(line):
                    li = line.strip()
                    if not li.startswith('#'):
                        match = resource_decorator.parse_resources(line)

                        if match is not None:
                            req_res_str = match.group(1)
                            if req_res_str is not None:
                                req_res = [y.strip() for y in req_res_str.split(',')]

                                for res_str in req_res:
                                    try:
                                        res = res_str.split("=")

                                        if res[0] == 'memory':
                                            memory = max(int(res[1]), memory) # Can be multiple decorators per flow
                                            logger.debug(f"Detected resource request for memory: {memory}")
                                            check_resources = True
                                        elif res[0] == 'cpu':
                                            cpu = max(int(res[1]), cpu) # Can be multiple decorators per flow
                                            logger.debug(f"Detected resource request for cpu: {cpu}")
                                            check_resources = True
                                        else:
                                            logger.debug(f"Detected request for unknown resource")
                                            return True
                                    except Exception as e:
                                        logger.debug(f"Error parsing resources")
                                        return True

        if check_resources:
            num_cpu = psutil.cpu_count()
            free_mem = psutil.virtual_memory().free >> 20 # shift to MBs, think this is correct?

            if cpu != 0:
                logger.debug(f"Requested {cpu} cpus, system has {num_cpu}")

            if memory != 0:
                logger.debug(f"Requested memory is {memory}, system has {free_mem} free")

            if cpu != 0 and num_cpu < int(cpu):
                return False

            if memory != 0 and free_mem < int(memory):
                return False

        return True



import requests
import os
import json
import logging
import boto3

from hashlib import sha256
from pathlib import Path
from enum import Enum, unique
from ..utils import ROOT_DIR

logger = logging.getLogger(__name__)


@unique
class ComputeStatus(Enum):
    NO_VALUE = 0
    INIT = 1
    CREATING = 2
    CREATED = 3
    FAILED = 4


class UserNotFound(Exception):
    ...


class IncorrectPassword(Exception):
    ...

class FreeTrialEnded(Exception):
    ...

class ProvisionError(Exception):

    def __init__(self, message: str, code: int):
        super().__init__(message)
        self.code = code


class Account:

    JSON_CONFIG = json.load(open(os.path.join(ROOT_DIR, "config.json"), "r"))
    BASE_URL = JSON_CONFIG['BASE_API_URL']
    CLIENT_ID = JSON_CONFIG['USER_POOL_CLIENT_ID']
    PROVISION_URL = f"{BASE_URL}/compute/config"

    def __init__(self, email: str, password: str):

        self._email = email
        self._password = password
        self._auth_token = None
        self._id_token = None
        self._user_home = str(Path.home())

    def login(self):

        cognito = boto3.client('cognito-idp', region_name='eu-west-1')

        try:
            logger.debug("Account::login()")
            response = cognito.initiate_auth(
                AuthFlow="USER_PASSWORD_AUTH",
                AuthParameters={
                    "USERNAME": self._email,
                    "PASSWORD": self._password
                },
                ClientId=self.CLIENT_ID
            )

        except cognito.exceptions.UserNotFoundException:
            logger.debug('User Not Found')
            raise UserNotFound
        except cognito.exceptions.NotAuthorizedException as e:
            if 'User is disabled' in str(e):
                raise FreeTrialEnded
            logger.debug('Incorrect Password')
            raise IncorrectPassword
        except Exception as e:
            logger.debug('Client Error: {e}')
            raise

        logger.debug(f'initiate_auth RESPONSE: {response}')

        if 'AuthenticationResult' not in response:
            raise Exception("Error logging in")

        logger.debug(
            f"Auth Token: {response['AuthenticationResult']['AccessToken']}")

        self._auth_token = response['AuthenticationResult']['AccessToken']
        self._refresh_token = response['AuthenticationResult']['RefreshToken']

    def provision(self, is_first_provision: bool) -> ComputeStatus:

        data = dict(
            company="solo",
            project="default"
        )

        logger.debug(f"Account::provision(): {data}")

        r = requests.post(self.PROVISION_URL,
                          json=data,
                          headers={
                              "x-api-key": self._auth_token
                          }
                          )

        logger.debug(f'Provision Status Code: {r.status_code}')

        if r.status_code == 202:
            return ComputeStatus.INIT
        elif r.status_code == 207:
            return ComputeStatus.CREATING
        elif r.status_code != 200:
            logger.debug(r.text)
            raise ProvisionError(r.text, r.status_code)

        response_data = r.json()

        if is_first_provision:
            self.login()

        response_data['config']['AERO_REFRESH_TOKEN'] = self._refresh_token
        response_data['config']['USER_POOL_CLIENT_ID'] = self.CLIENT_ID

        # Create Metaflow directory
        if not os.path.exists(f"{self._user_home}/.metaflowconfig"):
            os.mkdir(f"{self._user_home}/.metaflowconfig")

        with open(f"{self._user_home}/.metaflowconfig/config.json", 'w', encoding='utf-8') as f:
            json.dump(response_data['config'], f, ensure_ascii=False, indent=4)

        return ComputeStatus.CREATED

    @staticmethod
    def _hash(password):

        return sha256(password.strip()).hexdigest()



import setuptools
import re

from setuptools.command.develop import develop
from setuptools.command.install import install


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        print('MH_TEST_PostDevelopCommand')


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        print('MH_TEST_PostInstallCommand')


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

VERSIONFILE="aeroplatform/version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setuptools.setup(
    name="aeroplatform",
    version=verstr,
    author="Aero Technologies",
    author_email="aero@robbiea.co.uk",
    description="A simple Data Infrastructure Platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'click>=7.0',
        'requests',
        'boto3',
        'pylint<2.5.0',
        'aero-metaflow',
        'rich',
        'psutil'
    ],
    entry_points={
        'console_scripts': [
            'aero = aeroplatform.cli:cli',
        ],
    },
    include_package_data=True,
    package_data={
        "aeroplatform": [
            "config.json"
        ],
    },
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    }
)
